from z3c.relationfield.event import realize_relations
from z3c.relationfield.index import RelationCatalog
from z3c.relationfield.relation import RelationValue
from z3c.relationfield.relation import TemporaryRelationValue
from z3c.relationfield.relation import create_relation
from z3c.relationfield.schema import Relation
from z3c.relationfield.schema import RelationChoice
from z3c.relationfield.schema import RelationList



from functools import total_ordering

from persistent import Persistent
from z3c.objpath.interfaces import IObjectPath
from zope import component
from zope.interface import Declaration
from zope.interface import implementer
from zope.interface import providedBy
from zope.intid.interfaces import IIntIds

from z3c.relationfield.interfaces import IRelationValue
from z3c.relationfield.interfaces import ITemporaryRelationValue


@implementer(IRelationValue)
@total_ordering
class RelationValue(Persistent):

    _broken_to_path = None

    def __init__(self, to_id):
        self.to_id = to_id
        # these will be set automatically by events
        self.from_object = None
        self.__parent__ = None
        self.from_attribute = None

    def __hash__(self):
        """There should only be one RelationVaule per sort_key."""
        return id(self) // 16

    @property
    def from_id(self):
        intids = component.getUtility(IIntIds)
        return intids.getId(self.from_object)

    @property
    def from_path(self):
        return _path(self.from_object)

    @property
    def from_interfaces(self):
        return providedBy(self.from_object)

    @property
    def from_interfaces_flattened(self):
        return _interfaces_flattened(self.from_interfaces)

    @property
    def to_object(self):
        return _object(self.to_id)

    @property
    def to_path(self):
        if self.to_object is None:
            return self._broken_to_path
        return _path(self.to_object)

    @property
    def to_interfaces(self):
        return providedBy(self.to_object)

    @property
    def to_interfaces_flattened(self):
        return _interfaces_flattened(self.to_interfaces)

    def __eq__(self, other):
        if not isinstance(other, RelationValue):
            return False
        self_sort_key = self._sort_key()
        other_sort_key = other._sort_key()
        # if one of the relations we are comparing doesn't have a source
        # yet, only compare targets. This is to make comparisons within
        # ChoiceWidget work; a stored relation would otherwise not compare
        # equal with a relation generated for presentation in the UI
        if self_sort_key[0] is None or other_sort_key[0] is None:
            return self_sort_key[-1] == other_sort_key[-1]
        # otherwise do a full comparison
        return self_sort_key == other_sort_key

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        """Relations are sorted by default on a combination of the relation
        name, the path of the object the relation is one and the path of the
        object the relation is pointing to.
        """
        if (self.from_attribute or '') < (other.from_attribute or ''):
            return True
        if (self.from_path or '') < (other.from_path or ''):
            return True
        if (self.to_path or '') < (other.to_path or ''):
            return True
        return False

    def _sort_key(self):
        return (self.from_attribute, self.from_path, self.to_path)

    def broken(self, to_path):
        self._broken_to_path = to_path
        self.to_id = None

    def isBroken(self):
        return self.to_id is None


@implementer(ITemporaryRelationValue)
class TemporaryRelationValue(Persistent):
    """A relation that isn't fully formed yet.

    It needs to be finalized afterwards, when we are sure all potential
    target objects exist.
    """

    def __init__(self, to_path):
        self.to_path = to_path

    def convert(self):
        return create_relation(self.to_path)


def _object(id):
    if id is None:
        return None
    intids = component.getUtility(IIntIds)
    try:
        return intids.getObject(id)
    except KeyError:
        # XXX catching this error is not the right thing to do.
        # instead, breaking a relation by removing an object should
        # be caught and the relation should be adjusted that way.
        return None


def _path(obj):
    if obj is None:
        return ''
    object_path = component.getUtility(IObjectPath)
    return object_path.path(obj)


def _interfaces_flattened(interfaces):
    return Declaration(*interfaces).flattened()


def create_relation(to_path):
    """Create a relation to a particular path.

    Will create a broken relation if the path cannot be resolved.
    """
    object_path = component.getUtility(IObjectPath)
    try:
        to_object = object_path.resolve(to_path)
        intids = component.getUtility(IIntIds)
        return RelationValue(intids.getId(to_object))
    except ValueError:
        # create broken relation
        result = RelationValue(None)
        result.broken(to_path)
        return result



from lxml import etree
from z3c.schema2xml import IXMLGenerator
from zope.component import adapter
from zope.interface import implementer

from z3c.relationfield.interfaces import IRelation
from z3c.relationfield.interfaces import IRelationList
from z3c.relationfield.relation import TemporaryRelationValue


@adapter(IRelationList)
@implementer(IXMLGenerator)
class RelationListGenerator:
    """Export a relation list to XML.
    """

    def __init__(self, context):
        self.context = context

    def output(self, container, value):
        element = etree.SubElement(container, self.context.__name__)
        field = self.context.value_type
        if value is not None:
            for v in value:
                IXMLGenerator(field).output(element, v)

    def input(self, element):
        field = self.context.value_type
        return [
            IXMLGenerator(field).input(sub_element)
            for sub_element in element]


@adapter(IRelation)
@implementer(IXMLGenerator)
class RelationGenerator:
    """Eport a relation to XML.
    """

    def __init__(self, context):
        self.context = context

    def output(self, container, value):
        element = etree.SubElement(container, self.context.__name__)
        if value is not None:
            element.text = value.to_path

    def input(self, element):
        if element.text is None:
            return None
        path = element.text
        return TemporaryRelationValue(path)



from unittest import TestCase
from unittest import TestSuite
from unittest import defaultTestLoader

from zope.interface.interfaces import ComponentLookupError
from zope.interface.interfaces import ObjectEvent

from z3c.relationfield import Relation
from z3c.relationfield import RelationList
from z3c.relationfield import event
from z3c.relationfield.testing import MockContent
from z3c.relationfield.testing import register_fake_catalog
from z3c.relationfield.testing import register_fake_intid
from z3c.relationfield.testing import unregister_fake_catalog
from z3c.relationfield.testing import unregister_fake_intid


class FieldTests(TestCase):
    """Unit tests for fields
    """

    def test_list_value_type(self):
        f = RelationList(title="Test")
        self.assertTrue(isinstance(f.value_type, Relation))


class EventTests(TestCase):
    """Unit tests for the relation field event handlers.  The event handlers
    should be tolerant of missing utilities and objects not yet registered with
    the intid utility."""

    def setUp(self):
        self.content = MockContent()
        register_fake_catalog()
        register_fake_intid()

    def tearDown(self):
        unregister_fake_catalog()
        unregister_fake_intid()

    def test_missing_intids(self):
        """Event handlers which trigger on object removal should not
        throw exceptions when the utilities are missing.  The utilities may
        have been deleted in the same transaction (e.g. site deletion)."""
        # Remove intid utility and ensure the event handler doesn't fail
        unregister_fake_intid()
        try:
            event.breakRelations(ObjectEvent(self.content))
        except ComponentLookupError:
            self.fail("breakRelations fails when intid utility is missing")

    def test_break_relations_missing_catalog(self):
        # Remove relations catalog and ensure the event handler doesn't fail
        unregister_fake_catalog()
        try:
            event.breakRelations(ObjectEvent(self.content))
        except ComponentLookupError:
            self.fail("breakRelations fails when catalog utility is missing")

    def test_remove_relations_missing_catalog(self):
        # Remove relations catalog and ensure the event handler doesn't fail
        unregister_fake_catalog()
        try:
            event.removeRelations(self.content, None)
        except ComponentLookupError:
            self.fail("removeRelations fails when catalog utility is missing")

    def test_initid_failure(self):
        """When an object has not yet been registered with the intid utility,
        getId fails with a KeyError.  The event handlers need to handle this
        cleanly."""
        # Our content is not yet registered with the intid utility
        # and has no __parent__ attribute
        try:
            event.updateRelations(self.content, None)
        except (KeyError, AttributeError):
            self.fail("updateRelations fails with unregistered object")


def test_suite():
    suite = TestSuite()
    suite.addTest(defaultTestLoader.loadTestsFromTestCase(EventTests))
    suite.addTest(defaultTestLoader.loadTestsFromTestCase(FieldTests))
    return suite



import doctest
import sys
import unittest

from zope.component.testlayer import ZCMLFileLayer


# Evil hack to make pickling work with classes defined in doc tests
class NoCopyDict(dict):
    def copy(self):
        return self


class FakeModule:
    """A fake module."""

    def __init__(self, dict):
        self.__dict = dict

    def __getattr__(self, name):
        try:
            return self.__dict[name]
        except KeyError:
            raise AttributeError(name)


def setUp(test):
    test.globs['__name__'] = 'builtins'
    test.globs = NoCopyDict(test.globs)
    sys.modules['builtins'] = FakeModule(test.globs)


def tearDown(test):
    del sys.modules['builtins']
    test.globs.clear()


def test_suite():
    globs = {}
    readme = doctest.DocFileSuite(
        'README.rst',
        globs=globs,
        setUp=setUp,
        tearDown=tearDown,
        optionflags=doctest.ELLIPSIS,
    )
    import z3c.relationfield
    readme.layer = ZCMLFileLayer(z3c.relationfield)
    return unittest.TestSuite([readme])



import os

from setuptools import find_packages
from setuptools import setup


def read(*rnames):
    return open(os.path.join(*rnames)).read()


long_description = (
    read('src', 'z3c', 'relationfield', 'README.rst')
    + '\n' +
    read('CHANGES.rst')
    + '\n' +
    'Download\n'
    '********\n'
)

setup(
    name='z3c.relationfield',
    version='2.0.0',
    description="A relation field framework for Zope 3.",
    long_description=long_description,
    classifiers=[
        'Development Status :: 6 - Mature',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
    keywords='relation field',
    author='Martijn Faassen',
    author_email='zope-dev@zope.dev',
    url='https://github.com/zopefoundation/z3c.relationfield',
    license='ZPL 2.1',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['z3c'],
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.8',
    install_requires=[
        'setuptools',
        'z3c.objpath',
        'zc.relation >= 1.0',
        'zope.intid',
    ],
    extras_require={
        'test': [
            'zope.container',
            'zope.copypastemove',
            'zope.site',
        ],
        'xml': ['z3c.schema2xml >= 1.0',
                'lxml'],
    },
    entry_points={},
)



import BTrees
from zc.relation.catalog import Catalog
from zope import component
from zope.intid.interfaces import IIntIds

from z3c.relationfield.interfaces import IRelationValue


DEFAULT_INDEXES = [
    {
        'element': IRelationValue['from_id'],
    },
    {
        'element': IRelationValue['to_id'],
    },
    {
        'element': IRelationValue['from_attribute'],
        'kwargs': {
            'btree': BTrees.family32.OI,
        },
    },
    {
        'element': IRelationValue['from_interfaces_flattened'],
        'kwargs': {
            'btree': BTrees.family32.OI,
            'multiple': True,
        },
    },
    {
        'element': IRelationValue['to_interfaces_flattened'],
        'kwargs': {
            'btree': BTrees.family32.OI,
            'multiple': True,
        },
    },
]


def dump(obj, catalog, cache):
    intids = cache.get('intids')
    if intids is None:
        intids = cache['intids'] = component.getUtility(IIntIds)
    return intids.getId(obj)


def load(token, catalog, cache):
    intids = cache.get('intids')
    if intids is None:
        intids = cache['intids'] = component.getUtility(IIntIds)
    return intids.getObject(token)


class RelationCatalog(Catalog):

    def __init__(self, indexes=DEFAULT_INDEXES):
        """Initialize the catalog with indexes.

        Uses defaults if no special configuration was passed.
        """
        Catalog.__init__(self, dump, load)
        for index in indexes:
            self.addValueIndex(index['element'], **index.get('kwargs', {}))
